#include <stdio.h>
#include <sys/mman.h>
#include <pthread.h>
#include <linux/kvm.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/ioctl.h>
#define _GNU_SOURCE
#include <unistd.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <string.h>

#define KVM_DEVICE "/dev/kvm"
#define RAM_SIZE 512000000
#define CODE_START 0x1000
#define BINARY_FILE1 "guest1.bin"
#define BINARY_FILE2 "guest2.bin"
#define CURRENT_TIME ((double)clock() / CLOCKS_PER_SEC)

// FRAC_A + FRAC_B = 10
#define FRAC_A 8 // can be any number between 0 and 10
#define FRAC_B 2 // can be any number between 0 and 10
#define QUANTUM 1 // in seconds

struct vm
{
    int dev_fd;
    int kvm_version;
    int vm_fd;
    struct kvm_userspace_memory_region mem;
    struct vcpu *vcpus;
    __u64 ram_size;
    __u64 ram_start;
    int vcpu_number;
};

struct vcpu
{
    int vcpu_id;
    int vcpu_fd;
    pthread_t vcpu_thread;
    struct kvm_run *kvm_run;
    int kvm_run_mmap_size;
    struct kvm_regs regs;
    struct kvm_sregs sregs;
    void *(*vcpu_thread_func)(void *);
};

void kvm_init(struct vm *vm1, struct vm *vm2)
{
    int dev_fd = open(KVM_DEVICE, O_RDWR);

    if (dev_fd < 0)
    {
        perror("open /dev/kvm");
        exit(1);
    }

    int kvm_version = ioctl(dev_fd, KVM_GET_API_VERSION, 0);

    if (kvm_version < 0)
    {
        perror("KVM_GET_API_VERSION");
        exit(1);
    }

    if (kvm_version != KVM_API_VERSION)
    {
        fprintf(stderr, "Got KVM api version %d, expected %d\n",
                kvm_version, KVM_API_VERSION);
        exit(1);
    }

    vm1->dev_fd = dev_fd;
    vm2->dev_fd = dev_fd;
    vm1->kvm_version = kvm_version;
    vm2->kvm_version = kvm_version;
}

int kvm_create_vm(struct vm *vm, int ram_size)
{
    int ret = 0;
    vm->vm_fd = ioctl(vm->dev_fd, KVM_CREATE_VM, 0);

    if (vm->vm_fd < 0)
    {
        perror("can not create vm");
        return -1;
    }

    vm->ram_size = ram_size;
    vm->ram_start = (__u64)mmap(NULL, vm->ram_size,
                                PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_NORESERVE,
                                -1, 0);

    if ((void *)vm->ram_start == MAP_FAILED)
    {
        perror("can not mmap ram");
        return -1;
    }

    vm->mem.slot = 0;
    vm->mem.guest_phys_addr = 0;
    vm->mem.memory_size = vm->ram_size;
    vm->mem.userspace_addr = vm->ram_start;

    ret = ioctl(vm->vm_fd, KVM_SET_USER_MEMORY_REGION, &(vm->mem));

    if (ret < 0)
    {
        perror("can not set user memory region");
        return ret;
    }

    return ret;
}

void load_binary(struct vm *vm, char *binary_file)
{
    int fd = open(binary_file, O_RDONLY);

    if (fd < 0)
    {
        fprintf(stderr, "can not open binary file\n");
        exit(1);
    }

    int ret = 0;
    char *p = (char *)vm->ram_start;

    while (1)
    {
        ret = read(fd, p, 4096);
        if (ret <= 0)
        {
            break;
        }
        printf("VMFD: %d, Loaded Program with size: %d\n", vm->vm_fd, ret);
        p += ret;
    }
}

struct vcpu *kvm_init_vcpu(struct vm *vm, int vcpu_id)
{
    struct vcpu *vcpu = malloc(sizeof(struct vcpu));
    vcpu->vcpu_id = vcpu_id;
    vcpu->vcpu_fd = ioctl(vm->vm_fd, KVM_CREATE_VCPU, vcpu->vcpu_id);

    if (vcpu->vcpu_fd < 0)
    {
        perror("can not create vcpu");
        return NULL;
    }

    vcpu->kvm_run_mmap_size = ioctl(vm->dev_fd, KVM_GET_VCPU_MMAP_SIZE, 0);

    if (vcpu->kvm_run_mmap_size < 0)
    {
        perror("can not get vcpu mmsize");
        return NULL;
    }

    vcpu->kvm_run = mmap(NULL, vcpu->kvm_run_mmap_size, PROT_READ | PROT_WRITE, MAP_SHARED, vcpu->vcpu_fd, 0);

    if (vcpu->kvm_run == MAP_FAILED)
    {
        perror("can not mmap kvm_run");
        return NULL;
    }
    return vcpu;
}

void kvm_reset_vcpu(struct vcpu *vcpu)
{
    if (ioctl(vcpu->vcpu_fd, KVM_GET_SREGS, &(vcpu->sregs)) < 0)
    {
        perror("can not get sregs\n");
        exit(1);
    }

    vcpu->sregs.cs.selector = CODE_START;
    vcpu->sregs.cs.base = CODE_START * 16;
    vcpu->sregs.ss.selector = CODE_START;
    vcpu->sregs.ss.base = CODE_START * 16;
    vcpu->sregs.ds.selector = CODE_START;
    vcpu->sregs.ds.base = CODE_START * 16;
    vcpu->sregs.es.selector = CODE_START;
    vcpu->sregs.es.base = CODE_START * 16;
    vcpu->sregs.fs.selector = CODE_START;
    vcpu->sregs.fs.base = CODE_START * 16;
    vcpu->sregs.gs.selector = CODE_START;

    if (ioctl(vcpu->vcpu_fd, KVM_SET_SREGS, &vcpu->sregs) < 0)
    {
        perror("can not set sregs");
        exit(1);
    }

    vcpu->regs.rflags = 0x0000000000000002ULL;
    vcpu->regs.rip = 0;
    vcpu->regs.rsp = 0xffffffff;
    vcpu->regs.rbp = 0;

    if (ioctl(vcpu->vcpu_fd, KVM_SET_REGS, &(vcpu->regs)) < 0)
    {
        perror("KVM SET REGS\n");
        exit(1);
    }
}

/**
 * @brief Custom signal handler to clear the signal
 * Although this can be used but is incorrect as it does not handle corner cases.
 * Commented out as it is not used.
 * 
 */
// void clear(int sig, siginfo_t *siginfo, void *ran)
// {
//     // do nothing
// }

void kvm_run_vm(struct vm *vm1, struct vm *vm2)
{
    int ret;
    timer_t timer_id;

    // set up the sigevent and timer to send a signal periodically
    struct sigevent sevent;
    sevent.sigev_notify = SIGEV_SIGNAL;
    sevent.sigev_value.sival_ptr = &timer_id;
    sevent.sigev_signo = SIGUSR1; // any signals except for SIGKILL and SIGSTOP

    if (timer_create(CLOCK_MONOTONIC, &sevent, &timer_id) < 0)
    {
        perror("timer_create failure");
        exit(1);
    }

    /* Solutions with custom signal handler (works but incorrect [corner cases are not handled]) */
    // struct sigaction action;
    // action.sa_sigaction = clear;
    // sigemptyset(&action.sa_mask);
    // sigaction(SIGUSR1, &action, NULL);

    // set up the signal mask to block the signal
    sigset_t sig;
    sigemptyset(&sig);
    sigaddset(&sig, SIGUSR1);

    // block the signal and save the old signal mask
    sigset_t old_sig;
    sigprocmask(SIG_BLOCK, &sig, &old_sig);

    // set up the signal mask for the VMs
    // basically the signal will be unblocked for the VMs
    struct kvm_signal_mask *kvm_signal_mask = malloc(sizeof(struct kvm_signal_mask) + sizeof(unsigned long));
    kvm_signal_mask->len = sizeof(unsigned long);
    if (kvm_signal_mask->len > sizeof(old_sig))
    {
        perror("sigset broken");
        exit(1);
    }

    memcpy(kvm_signal_mask->sigset, &old_sig, kvm_signal_mask->len);

    if (ioctl(vm1->vcpus->vcpu_fd, KVM_SET_SIGNAL_MASK, kvm_signal_mask) < 0)
    {
        perror("KVM_SETSIGNAL_MASK failure");
        exit(1);
    }
    if (ioctl(vm2->vcpus->vcpu_fd, KVM_SET_SIGNAL_MASK, kvm_signal_mask) < 0)
    {
        perror("KVM_SETSIGNAL_MASK failure");
        exit(1);
    }

    // structures for timer set and reset
    struct itimerspec timer_set = {
        .it_value.tv_sec = QUANTUM,
        .it_value.tv_nsec = 0,
    };

    struct timespec timer_reset = {
        .tv_sec = 0,
        .tv_nsec = 0,
    };

    // set the timer to kick the virtual machine after the quantum
    if (timer_settime(timer_id, 0, &timer_set, NULL) < 0)
    {
        perror("timer_settime failure");
        exit(1);
    }

    // setup the segmenation and other vcpu things
    kvm_reset_vcpu(vm1->vcpus);
    kvm_reset_vcpu(vm2->vcpus);

    // set the budget for the VMs
    int budget_vm1 = FRAC_A;
    int budget_vm2 = FRAC_B;

    // sanity check for the fractions
    if (budget_vm1 + budget_vm2 != 10)
    {
        perror("fractions are not correct");
        exit(1);
    }

    struct vm *vm;
    int turn = 0;

    while (1)
    {
        // Run alternate VMs in round robin order until budget is exhausted
        // when it is exhausted, reset the budget and continue
        // Note: This is a simple round robin scheduler not best but works
        if (turn == 0 && budget_vm1 > 0)
        {
            vm = vm1;
            budget_vm1--;
            turn = 1;
        }
        else if (turn == 1 && budget_vm2 > 0)
        {
            vm = vm2;
            budget_vm2--;
            turn = 0;
        }
        else if (budget_vm1 > 0)
        {
            vm = vm1;
            budget_vm1--;
            turn = 1;
        }
        else if (budget_vm2 > 0)
        {
            vm = vm2;
            budget_vm2--;
            turn = 0;
        }
        else if (budget_vm1 == 0 && budget_vm2 == 0)
        {
            budget_vm1 = FRAC_A;
            budget_vm2 = FRAC_B;
            continue;
        }
        else
        {
            perror("Something went wrong");
            exit(1);
        }

        printf("VMFD: %d started running\n", vm->vm_fd);
        ret = ioctl(vm->vcpus->vcpu_fd, KVM_RUN, 0);

        // Get rax register value for sanity check
        ret = ioctl(vm->vcpus->vcpu_fd, KVM_GET_REGS, &(vm->vcpus->regs));
        if (ret < 0)
        {
            perror("KVM_GET_REGS failed");
            exit(1);
        }

        printf("VMFD: %d stopped running - exit reason: %d, rax: %lld\n", vm->vm_fd, vm->vcpus->kvm_run->exit_reason, vm->vcpus->regs.rax);

        switch (vm->vcpus->kvm_run->exit_reason)
        {
        case KVM_EXIT_UNKNOWN:
            printf("VMFD: %d KVM_EXIT_UNKNOWN\n", vm->vm_fd);
            break;
        case KVM_EXIT_DEBUG:
            printf("VMFD: %d KVM_EXIT_DEBUG\n", vm->vm_fd);
            break;
        case KVM_EXIT_IO:
            printf("VMFD: %d KVM_EXIT_IO\n", vm->vm_fd);
            printf("VMFD: %d out port: %d, data: %d\n", vm->vm_fd, vm->vcpus->kvm_run->io.port, *(int *)((char *)(vm->vcpus->kvm_run) + vm->vcpus->kvm_run->io.data_offset));
            sleep(1);
            break;
        case KVM_EXIT_MMIO:
            printf("VMFD: %d KVM_EXIT_MMIO\n", vm->vm_fd);
            break;
        case KVM_EXIT_INTR:
            printf("VMFD: %d KVM_EXIT_INTR\n", vm->vm_fd);

            // clear the signal and reset the timer
            int ret = sigtimedwait(&sig, NULL, &timer_reset);
            if (ret < 0 && errno != EAGAIN)
            {
                perror("sigtimedwait failure");
                exit(1);
            }

            // set the timer to kick the virtual machine after the quantum
            if (timer_settime(timer_id, 0, &timer_set, NULL) < 0)
            {
                perror("timer_settime failure");
                exit(1);
            }

            /* Solutions with custom signal handler (works but incorrect [corner cases are not handled]) */
            // if (timer_settime(timer_id, 0, &timeout, NULL) < 0)
            // {
            //     perror("timer_settime failure");
            //     exit(1);
            // }
            break;
        case KVM_EXIT_SHUTDOWN:
            printf("VMFD: %d KVM_EXIT_SHUTDOWN\n", vm->vm_fd);
            goto exit_kvm;
            break;
        default:
            printf("VMFD: %d KVM PANIC\n", vm->vm_fd);
            printf("VMFD: %d KVM exit reason: %d\n", vm->vm_fd, vm->vcpus->kvm_run->exit_reason);
            goto exit_kvm;
        }

        if (ret < 0 && vm->vcpus->kvm_run->exit_reason != KVM_EXIT_INTR)
        {
            fprintf(stderr, "VMFD: %d KVM_RUN failed\n", vm->vm_fd);
            printf("VMFD: %d KVM_RUN return value %d\n", vm->vm_fd, ret);
            exit(1);
        }
    }

exit_kvm:
    exit(0);
}

void kvm_clean_vm(struct vm *vm)
{
    close(vm->vm_fd);
    munmap((void *)vm->ram_start, vm->ram_size);
}

void kvm_clean_vcpu(struct vcpu *vcpu)
{
    munmap(vcpu->kvm_run, vcpu->kvm_run_mmap_size);
    close(vcpu->vcpu_fd);
}

void kvm_clean(struct vm *vm)
{
    assert(vm != NULL);
    close(vm->dev_fd);
    free(vm);
}

int main(int argc, char **argv)
{
    struct vm *vm1 = malloc(sizeof(struct vm));
    struct vm *vm2 = malloc(sizeof(struct vm));

    kvm_init(vm1, vm2);

    if (kvm_create_vm(vm1, RAM_SIZE) < 0)
    {
        fprintf(stderr, "create vm fault\n");
        return -1;
    }

    if (kvm_create_vm(vm2, RAM_SIZE) < 0)
    {
        fprintf(stderr, "create vm fault\n");
        return -1;
    }

    load_binary(vm1, BINARY_FILE1);
    load_binary(vm2, BINARY_FILE2);

    vm1->vcpu_number = 1;
    vm1->vcpus = kvm_init_vcpu(vm1, 0);

    vm2->vcpu_number = 1;
    vm2->vcpus = kvm_init_vcpu(vm2, 0);

    kvm_run_vm(vm1, vm2);

    kvm_clean_vm(vm1);
    kvm_clean_vm(vm2);

    kvm_clean_vcpu(vm1->vcpus);
    kvm_clean_vcpu(vm2->vcpus);
    kvm_clean(vm1);
    kvm_clean(vm2);
}